package net.knarcraft.knarlib.formatting;

import org.jetbrains.annotations.NotNull;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import static net.knarcraft.knarlib.formatting.StringFormatter.replacePlaceholder;

/**
 * A utility for formatting a string specifying an amount of time
 */
@SuppressWarnings("unused")
public final class TimeFormatter {

    private static Map<Double, TranslatableTimeUnit[]> timeUnits;
    private static List<Double> sortedUnits;

    private TimeFormatter() {

    }

    /**
     * Gets the string used for displaying this sign's duration
     *
     * @param translator <p>The translator to use for translation</p>
     * @param duration   <p>The duration, in seconds, to display</p>
     * @return <p>The string used for displaying this sign's duration</p>
     */
    public static String getDurationString(@NotNull Translator translator, long duration) {
        //Initialize time units once, and only if this method is actually used
        if (sortedUnits == null) {
            initializeUnits(translator);
        }

        //This case is handled separately to avoid a divide by 0 exception
        if (duration == 0) {
            return translator.getTranslatedMessage(TranslatableTimeUnit.UNIT_NOW);
        }

        //Loop through all units to find the largest unit less than the remaining duration
        for (double unit : sortedUnits) {
            if (duration / unit < 1) {
                continue;
            }

            //Round the remaining units of time to two digits
            double units = round(duration / unit);
            //Cast to int if the number has no decimals
            boolean castToInt = (units * 10) % 10 == 0;
            //Select the singular form if the unit is 1, and the plural otherwise
            TranslatableMessage translatableUnit = timeUnits.get(unit)[units == 1 ? 0 : 1];
            return formatDurationString(units, translator, translatableUnit, castToInt);
        }
        //It can be assumed this case only happens between 0 and 1 seconds, thus plural is used
        return formatDurationString(duration, translator, TranslatableTimeUnit.UNIT_SECONDS, false);
    }

    /**
     * Rounds a number to its last two digits
     *
     * @param number <p>The number to round</p>
     * @return <p>The rounded number</p>
     */
    private static double round(double number) {
        return Math.round(number * 100.0) / 100.0;
    }

    /**
     * Formats a duration string
     *
     * @param duration            <p>The duration to display</p>
     * @param translator          <p>The translator to use for translation</p>
     * @param translatableMessage <p>The time unit to display</p>
     * @param castToInt           <p>Whether to cast the duration to an int</p>
     * @return <p>The formatted duration string</p>
     */
    private static String formatDurationString(double duration, @NotNull Translator translator,
                                               @NotNull TranslatableMessage translatableMessage, boolean castToInt) {
        String durationFormat = translator.getTranslatedMessage(TranslatableTimeUnit.DURATION_FORMAT);
        durationFormat = replacePlaceholder(durationFormat, "{unit}",
                translator.getTranslatedMessage(translatableMessage));
        return replacePlaceholder(durationFormat, "{time}", castToInt ? String.valueOf((int) duration) :
                String.valueOf(duration));
    }

    /**
     * Initializes the mapping of available time units for formatting permission sign duration
     */
    private static void initializeUnits(@NotNull Translator translator) {
        //Register each time unit with a registered singular and plural translation
        timeUnits = new HashMap<>();
        registerTimeUnit(translator, 315360000, TranslatableTimeUnit.UNIT_DECADE, TranslatableTimeUnit.UNIT_DECADES);
        registerTimeUnit(translator, 31536000, TranslatableTimeUnit.UNIT_YEAR, TranslatableTimeUnit.UNIT_YEARS);
        registerTimeUnit(translator, 2419200, TranslatableTimeUnit.UNIT_MONTH, TranslatableTimeUnit.UNIT_MONTHS);
        registerTimeUnit(translator, 604800, TranslatableTimeUnit.UNIT_WEEK, TranslatableTimeUnit.UNIT_WEEKS);
        registerTimeUnit(translator, 86400, TranslatableTimeUnit.UNIT_DAY, TranslatableTimeUnit.UNIT_DAYS);
        registerTimeUnit(translator, 3600, TranslatableTimeUnit.UNIT_HOUR, TranslatableTimeUnit.UNIT_HOURS);
        registerTimeUnit(translator, 60, TranslatableTimeUnit.UNIT_MINUTE, TranslatableTimeUnit.UNIT_MINUTES);
        boolean registeredSeconds = registerTimeUnit(translator, 1, TranslatableTimeUnit.UNIT_SECOND,
                TranslatableTimeUnit.UNIT_SECONDS);

        //As now and seconds are the base units, those must be registered to use getDurationString
        if (!registeredSeconds || !translator.hasTranslation(TranslatableTimeUnit.UNIT_NOW)) {
            throw new IllegalStateException("A translation is missing for UNIT_SECOND, UNIT_SECONDS or UNIT_NOW");
        }

        //Make a sorted list from each unit's number of seconds
        sortedUnits = new ArrayList<>(timeUnits.keySet());
        Collections.sort(sortedUnits);
        Collections.reverse(sortedUnits);
    }

    /**
     * Registers one time unit if its translations have been registered
     *
     * @param translator <p>The translator to use for checking whether necessary translations have been set</p>
     * @param seconds    <p>The amount of seconds the time unit corresponds to</p>
     * @param singular   <p>The translatable time unit to use for the singular form</p>
     * @param plural     <p>The translatable time unit to use for the plural form</p>
     * @return <p>True if the time unit was registered. False if a translation is missing</p>
     */
    private static boolean registerTimeUnit(@NotNull Translator translator, double seconds,
                                            @NotNull TranslatableTimeUnit singular,
                                            @NotNull TranslatableTimeUnit plural) {
        String singularTranslation = translator.getTranslatedMessage(singular);
        String pluralTranslation = translator.getTranslatedMessage(plural);
        if (translator.hasTranslation(singular) && translator.hasTranslation(plural)) {
            timeUnits.put(seconds, new TranslatableTimeUnit[]{singular, plural});
            return true;
        } else {
            return false;
        }
    }

}
