package net.knarcraft.knarlib.util;

import net.knarcraft.knarlib.particle.ParticleConfig;
import org.bukkit.Location;
import org.bukkit.World;
import org.bukkit.block.Block;
import org.bukkit.util.BoundingBox;
import org.bukkit.util.Vector;
import org.jetbrains.annotations.NotNull;

import java.util.HashMap;
import java.util.Map;
import java.util.UUID;

/**
 * A helper class for spawning particle effects
 */
@SuppressWarnings("unused")
public final class ParticleHelper {

    private static final Map<UUID, Vector[]> pyramidVectors = new HashMap<>();
    private static final Map<UUID, Double[][]> circleCoordinates = new HashMap<>();
    private static final Map<UUID, Double[][]> sphereCoordinates = new HashMap<>();

    private ParticleHelper() {

    }

    /**
     * Removes any stored calculations for the given id
     *
     * <p>If you frequently change the id without clearing, you'll create a memory leak!</p>
     *
     * @param id <p>The id specified when generating a shape</p>
     */
    public static void clearStoredCalculations(UUID id) {
        pyramidVectors.remove(id);
        circleCoordinates.remove(id);
        sphereCoordinates.remove(id);
    }

    /**
     * Spawns a cube of particles at the given location
     *
     * <p>It is recommended to only spawn one particle with no spread, and a density between 1 and 0.01</p>
     *
     * @param world          <p>The world to spawn the particles in</p>
     * @param location       <p>The location of the block to spawn the particles at</p>
     * @param particleConfig <p>The configuration describing the particle to spawn</p>
     * @param blockHeight    <p>The height of the block to spawn the particle above</p>
     */
    public static void drawCube(@NotNull World world, @NotNull Location location, @NotNull ParticleConfig particleConfig,
                                double blockHeight) {
        // Draw the top and bottom of the cube
        drawSquare(world, location, particleConfig, blockHeight);
        drawSquare(world, location.clone().add(0, 1, 0), particleConfig, blockHeight);

        for (float y = 0; y <= 1; y += particleConfig.getParticleDensity()) {
            double height = particleConfig.getHeightOffset() + y;
            spawnParticle(world, location.clone().add(0, height, 0), particleConfig, blockHeight);
            spawnParticle(world, location.clone().add(0, height, 1), particleConfig, blockHeight);
            spawnParticle(world, location.clone().add(1, height, 0), particleConfig, blockHeight);
            spawnParticle(world, location.clone().add(1, height, 1), particleConfig, blockHeight);
        }
    }

    /**
     * Spawns a sphere of particles at the given location
     *
     * <p>It is recommended to only spawn one particle with no spread, and a density between 4 and 0.1</p>
     *
     * @param world                <p>The world to spawn the particles in</p>
     * @param location             <p>The location of the block to spawn the particles at</p>
     * @param particleConfig       <p>The configuration describing the particle to spawn</p>
     * @param blockHeight          <p>The height of the block to spawn the particle above</p>
     * @param storedCalculationsId <p>An id for stored calculations. Use clearStoredCalculations with this id later.</p>
     */
    public static void drawSphere(@NotNull World world, @NotNull Location location, @NotNull ParticleConfig particleConfig,
                                  double blockHeight, UUID storedCalculationsId) {
        // For spheres, densities below 0.1 has weird bugs such as blinking in and out of existence, and floating point
        // errors when calculating the length of circleCoordinates
        double density = Math.max(1, particleConfig.getParticleDensity());
        // Store calculations for improved efficiency
        if (sphereCoordinates.get(storedCalculationsId) == null) {
            int length = (int) Math.ceil((180 / density));
            Double[][] coordinates = new Double[length * 3][];
            double height = particleConfig.getHeightOffset() + 0.5;
            int i = 0;
            for (float x = 0; x < 180; x += density) {
                if (i >= coordinates.length) {
                    continue;
                }

                double cos = 0.5 * Math.cos(x);
                double sin = 0.5 * Math.sin(x);
                coordinates[i++] = new Double[]{sin + 0.5, height, cos + 0.5};
                coordinates[i++] = new Double[]{sin + 0.5, height + cos, 0.5};
                coordinates[i++] = new Double[]{0.5, height + sin, cos + 0.5};
            }
            sphereCoordinates.put(storedCalculationsId, coordinates);
        }

        // Spawn particles on the stored locations, relative to the launchpad
        for (Double[] sphereCoordinate : sphereCoordinates.get(storedCalculationsId)) {
            spawnParticle(world, location.clone().add(sphereCoordinate[0], sphereCoordinate[1], sphereCoordinate[2]),
                    particleConfig, blockHeight);
        }
    }

    /**
     * Spawns a pyramid of particles at the given location
     *
     * <p>It is recommended to only spawn one particle with no spread, and a density between 1 and 0.01</p>
     *
     * @param world                <p>The world to spawn the particles in</p>
     * @param location             <p>The location of the block to spawn the particles at</p>
     * @param particleConfig       <p>The configuration describing the particle to spawn</p>
     * @param blockHeight          <p>The height of the block to spawn the particle above</p>
     * @param storedCalculationsId <p>An id for stored calculations. Use clearStoredCalculations with this id later.</p>
     */
    public static void drawPyramid(@NotNull World world, @NotNull Location location, @NotNull ParticleConfig particleConfig,
                                   double blockHeight, UUID storedCalculationsId) {
        // Draw the bottom of the pyramid
        drawSquare(world, location, particleConfig, blockHeight);

        // Store calculations for improved efficiency
        if (pyramidVectors.get(storedCalculationsId) == null) {
            // The 0.5 offsets are required for the angle of the pyramid's 4 lines to be correct
            double height = particleConfig.getHeightOffset();
            double coordinateMin = -0.5 * height;
            double coordinateMax = 1 + (0.5 * height);

            Vector[] vectors = new Vector[5];
            // The vector from the origin to the top of the pyramid
            vectors[0] = new Vector(0.5, 1 + height, 0.5);
            // The vectors from the top of the pyramid towards each corner
            vectors[1] = new Vector(coordinateMin, 0, coordinateMin).subtract(vectors[0]).normalize();
            vectors[2] = new Vector(coordinateMax, 0, coordinateMin).subtract(vectors[0]).normalize();
            vectors[3] = new Vector(coordinateMin, 0, coordinateMax).subtract(vectors[0]).normalize();
            vectors[4] = new Vector(coordinateMax, 0, coordinateMax).subtract(vectors[0]).normalize();
            pyramidVectors.put(storedCalculationsId, vectors);
        }

        Vector[] storedVectors = pyramidVectors.get(storedCalculationsId);
        Location topLocation = location.clone().add(storedVectors[0]);
        for (double x = 0; x <= 1.2; x += particleConfig.getParticleDensity()) {
            spawnParticle(world, topLocation.clone().add(storedVectors[1].clone().multiply(x)), particleConfig, blockHeight);
            spawnParticle(world, topLocation.clone().add(storedVectors[2].clone().multiply(x)), particleConfig, blockHeight);
            spawnParticle(world, topLocation.clone().add(storedVectors[3].clone().multiply(x)), particleConfig, blockHeight);
            spawnParticle(world, topLocation.clone().add(storedVectors[4].clone().multiply(x)), particleConfig, blockHeight);
        }
    }

    /**
     * Spawns a circle of particles at the given location
     *
     * <p>It is recommended to only spawn one particle with no spread, and a density between 4 and 0.1</p>
     *
     * @param world                <p>The world to spawn the particles in</p>
     * @param location             <p>The location of the block to spawn the particles at</p>
     * @param particleConfig       <p>The configuration describing the particle to spawn</p>
     * @param blockHeight          <p>The height of the block to spawn the particle above</p>
     * @param storedCalculationsId <p>An id for stored calculations. Use clearStoredCalculations with this id later.</p>
     */
    public static void drawCircle(@NotNull World world, @NotNull Location location, @NotNull ParticleConfig particleConfig,
                                  double blockHeight, UUID storedCalculationsId) {
        // For circles, densities below 0.1 has weird bugs such as blinking in and out of existence, and floating point
        // errors when calculating the length of circleCoordinates
        double density = Math.max(1, particleConfig.getParticleDensity());
        // Store calculations for improved efficiency
        if (circleCoordinates.get(storedCalculationsId) == null) {
            Double[][] coordinates = new Double[(int) Math.ceil((180 / density))][];
            int i = 0;
            for (float x = 0; x < 180; x += density) {
                if (i >= coordinates.length) {
                    continue;
                }
                coordinates[i++] = new Double[]{(0.5 * Math.sin(x)) + 0.5, (0.5 * Math.cos(x)) + 0.5};
            }
            circleCoordinates.put(storedCalculationsId, coordinates);
        }

        // Spawn particles on the stored locations, relative to the launchpad
        for (Double[] circleCoordinate : circleCoordinates.get(storedCalculationsId)) {
            spawnParticle(world, location.clone().add(circleCoordinate[0], particleConfig.getHeightOffset(),
                    circleCoordinate[1]), particleConfig, blockHeight);
        }
    }

    /**
     * Spawns a square of particles at the given location
     *
     * <p>It is recommended to only spawn one particle with no spread, and a density between 1 and 0.01</p>
     *
     * @param world          <p>The world to spawn the particles in</p>
     * @param location       <p>The location of the block to spawn the particles at</p>
     * @param particleConfig <p>The configuration describing the particle to spawn</p>
     * @param blockHeight    <p>The height of the block to spawn the particle above</p>
     */
    public static void drawSquare(@NotNull World world, @NotNull Location location, @NotNull ParticleConfig particleConfig,
                                  double blockHeight) {
        for (float x = 0; x <= 1; x += particleConfig.getParticleDensity()) {
            spawnParticle(world, location.clone().add(x, particleConfig.getHeightOffset(), 0), particleConfig, blockHeight);
            spawnParticle(world, location.clone().add(x, particleConfig.getHeightOffset(), 1), particleConfig, blockHeight);
            spawnParticle(world, location.clone().add(0, particleConfig.getHeightOffset(), x), particleConfig, blockHeight);
            spawnParticle(world, location.clone().add(1, particleConfig.getHeightOffset(), x), particleConfig, blockHeight);
        }
    }

    /**
     * Spawns the specified particle at the given location
     *
     * @param world          <p>The world to spawn the particle in</p>
     * @param location       <p>The location to spawn the particle at</p>
     * @param particleConfig <p>The configuration describing the particle to spawn</p>
     * @param blockHeight    <p>The height of the block to spawn the particle above</p>
     */
    public static void spawnParticle(@NotNull World world, @NotNull Location location,
                                     @NotNull ParticleConfig particleConfig, double blockHeight) {
        world.spawnParticle(particleConfig.getParticleType(),
                location.add(0, blockHeight, 0), particleConfig.getParticleAmount(),
                particleConfig.getOffsetX(), particleConfig.getOffsetY(), particleConfig.getOffsetZ(),
                particleConfig.getExtra());
    }

    /**
     * Gets the height of the block at the given location
     *
     * @param block <p>The block to check</p>
     * @return <p>The height of the block</p>
     */
    public static double getBlockHeight(@NotNull Block block) {
        double maxY = 0;
        for (BoundingBox boundingBox : block.getCollisionShape().getBoundingBoxes()) {
            if (boundingBox.getMaxY() > maxY) {
                maxY = boundingBox.getMaxY();
            }
        }
        return maxY;
    }

}
