package net.knarcraft.knarlib.util;

import org.bukkit.plugin.Plugin;
import org.bukkit.scheduler.BukkitScheduler;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.function.Consumer;
import java.util.function.Supplier;
import java.util.logging.Level;

/**
 * The update checker is responsible for looking for new updates
 */
@SuppressWarnings("unused")
public final class UpdateChecker {

    private final static String updateNotice = "A new update is available: %s (You are still on %s)";

    private UpdateChecker() {

    }

    /**
     * Checks if there's a new update available, and alerts the user if necessary
     *
     * @param plugin             <p>The plugin to check for updates for</p>
     * @param apiResourceURL     <p>The spigot URL to check for updates. Example:
     *                           <a href="https://api.spigotmc.org/legacy/update.php?resource=1111">
     *                           https://api.spigotmc.org/legacy/update.php?resource={resourceId}</a></p>
     * @param getVersionSupplier <p>The supplier used to get the current plugin version</p>
     * @param setVersionMethod   <p>A method to call with the new version as an argument. Can be used to alert admins about an available update or similar.</p>
     */
    public static void checkForUpdate(@NotNull Plugin plugin, @NotNull String apiResourceURL,
                                      @NotNull Supplier<String> getVersionSupplier,
                                      @NotNull Consumer<String> setVersionMethod) {
        BukkitScheduler scheduler = plugin.getServer().getScheduler();
        scheduler.runTaskAsynchronously(plugin, () -> UpdateChecker.queryAPI(plugin, apiResourceURL, getVersionSupplier,
                setVersionMethod));
    }

    /**
     * Queries the spigot API to check for a newer version, and prints to the console if found
     *
     * @param plugin           <p>The plugin to check for updates for</p>
     * @param APIResourceURL   <p>The spigot URL to check for updates</p>
     * @param getVersionMethod <p>The supplier used to get the current plugin version</p>
     * @param setVersionMethod <p>A method to call with the new version as an argument. Can be used to alert admins about an available update or similar.</p>
     */
    private static void queryAPI(@NotNull Plugin plugin, @NotNull String APIResourceURL,
                                 @NotNull Supplier<String> getVersionMethod,
                                 @Nullable Consumer<String> setVersionMethod) {
        try {
            InputStream inputStream = new URL(APIResourceURL).openStream();
            BufferedReader reader = FileHelper.getBufferedReaderFromInputStream(inputStream);
            //There should only be one line of output
            String newVersion = reader.readLine();
            reader.close();

            String oldVersion = getVersionMethod.get();
            //If there is a newer version, notify the user
            if (isVersionHigher(oldVersion, newVersion)) {
                plugin.getLogger().log(Level.WARNING, getUpdateAvailableString(newVersion, oldVersion));
                if (setVersionMethod != null) {
                    setVersionMethod.accept(newVersion);
                }
            }
        } catch (IOException e) {
            plugin.getLogger().log(Level.WARNING, "Unable to get the newest plugin version.");
        }
    }

    /**
     * Gets the string to display to a user to alert about a new update
     *
     * @param newVersion <p>The new available plugin version</p>
     * @param oldVersion <p>The old (current) plugin version</p>
     * @return <p>The string to display</p>
     */
    public static @NotNull String getUpdateAvailableString(@NotNull String newVersion, @NotNull String oldVersion) {
        return String.format(updateNotice, newVersion, oldVersion);
    }

    /**
     * Decides whether one version number is higher than another
     *
     * @param oldVersion <p>The old version to check</p>
     * @param newVersion <p>The new version to check</p>
     * @return <p>True if the new version is higher than the old one</p>
     */
    public static boolean isVersionHigher(@NotNull String oldVersion, @NotNull String newVersion) {
        oldVersion = removeNonNumericCharacters(oldVersion);
        newVersion = removeNonNumericCharacters(newVersion);
        String[] oldVersionParts = oldVersion.split("\\.");
        String[] newVersionParts = newVersion.split("\\.");
        int versionLength = Math.max(oldVersionParts.length, newVersionParts.length);
        for (int i = 0; i < versionLength; i++) {
            int oldVersionNumber = oldVersionParts.length > i ? Integer.parseInt(oldVersionParts[i]) : 0;
            int newVersionNumber = newVersionParts.length > i ? Integer.parseInt(newVersionParts[i]) : 0;
            if (newVersionNumber != oldVersionNumber) {
                return newVersionNumber > oldVersionNumber;
            }
        }
        return false;
    }

    /**
     * Removes all characters except numbers and the "." character from the given version string
     *
     * @param versionString <p>The version string to clean</p>
     * @return <p>The string with non-numeric characters replaced</p>
     */
    private static @NotNull String removeNonNumericCharacters(@NotNull String versionString) {
        return versionString.replaceAll("[^0-9.]", "");
    }

}
